//--------------------------------------------------------------------------------
//  Windows DIB data operations
//    Copyright (c) kei mesuda 2007 
//    (http://www.yomogi.sakura.ne.jp/~si)
//--------------------------------------------------------------------------------
private import std.stdio;
private import std.file;
private import util.color;




// Bitmap File/Info Header 
//--------------------------------------------------
private align(1) struct BFH {
    char[2] bfType = "BM";
    uint    bfSize = 0;
    ushort  bfReserved1 = 0;
    ushort  bfReserved2 = 0;
    uint    bfOffBits = 54;
}


private align(1) struct BIH {
    uint    biSize = 40;
    int     biWidth = 1;
    int     biHeight = 1;
    ushort  biPlanes = 1;
    ushort  biBitCount = 24;
    uint    biCompression = 0;
    uint    biSizeImage = 4;
    int     biXPelsPerMeter = 0;
    int     biYPelsPerMeter = 0;
    uint    biClrUsed = 0;
    uint    biClrImportant = 0;
}




public:
// load bmp file by specifyed bit size (8,24,32). 
ubyte[] loadBitmap(out ubyte[] pix_out, out int width, out int height, char[] nameFile, int bit, bool isRGBA)
{
    int bmpbit;
    scope ubyte[] pxl;
    scope ubyte[] tbl;
    
    readBitmap(nameFile, width, height, bmpbit, pxl, tbl);
    switch (bit) {
    case 8:  return convertBitmap8 (pix_out, pxl, tbl, bmpbit);
    case 24: return convertBitmap24(pix_out, pxl, tbl, bmpbit, isRGBA);
    case 32: return convertBitmap32(pix_out, pxl, tbl, bmpbit, isRGBA);
    default: throw new Exception(nameFile ~ "; bit size not supported");
    }
}




// read limited information of bitmap
ubyte[] readBitmap(char[] nameFile, out int width, out int height, out int bit, out ubyte[] pixel, out ubyte[] tableColor)
{
    ubyte[] data;
    try data = cast(ubyte[])std.file.read(nameFile);
    catch (Exception ex) throw new Exception("readBitmap(); " ~ nameFile ~ " does not exist");

    // read header
    if (data.length < 54) throw new Exception("readBitmap(); " ~ nameFile ~ " is not a bitmap file");
    ubyte[] header = data[0..14];
    ubyte[] info   = data[14..54];
    if (header[0]!='B' || header[1]!='M')    throw new Exception("readBitmap(); " ~ nameFile ~ " is not a bitmap file");
    if ((cast(BIH*)info).biCompression != 0) throw new Exception("readBitmap(); " ~ nameFile ~ " is compressed");

    // get information
    width  = (cast(BIH*)info).biWidth;
    height = (cast(BIH*)info).biHeight;
    bit    = (cast(BIH*)info).biBitCount;

    // refer color table
    uint cntColor = 0;
    if (bit==8 || bit==1) {
        cntColor = (cast(BIH*)info).biClrUsed;
        if (cntColor==0) cntColor=1<<bit;
        if (data.length < 54+cntColor) throw new Exception("readBitmap(); " ~ nameFile ~ " color table is broken");
        tableColor = data[54..54+(cntColor<<2)];
    }

    pixel = data[54+(cntColor<<2)..length];
    return data;
}




// write limited information of 24bit bitmap
ubyte[] writeBitmap(char[] nameFile, int width, int height, int pitch, ubyte[] pixel)
{
    ubyte[] buffer;
    int width_byte  = width * 3;
    int width_align = (width_byte + 3) & 0xfffffc;
    int size_image  = width_align * height;
    buffer.length = size_image + 54;
    ubyte[] data = buffer[54..$];
    
    (cast(BFH*)buffer[0..14]).bfType[] = "BM";
    (cast(BFH*)buffer[0..14]).bfReserved1 = 0;
    (cast(BFH*)buffer[0..14]).bfReserved2 = 0;
    (cast(BFH*)buffer[0..14]).bfOffBits = 54;
    
    (cast(BIH*)buffer[14..54]).biSize = 40;
    (cast(BIH*)buffer[14..54]).biPlanes = 1;
    (cast(BIH*)buffer[14..54]).biBitCount = 24;
    (cast(BIH*)buffer[14..54]).biCompression = 0;
    (cast(BIH*)buffer[14..54]).biXPelsPerMeter = 0;
    (cast(BIH*)buffer[14..54]).biYPelsPerMeter = 0;
    (cast(BIH*)buffer[14..54]).biClrUsed = 0;
    (cast(BIH*)buffer[14..54]).biClrImportant = 0;
    
    (cast(BFH*)buffer[0..14]).bfSize = buffer.length;
    (cast(BIH*)buffer[14..54]).biWidth = width;
    (cast(BIH*)buffer[14..54]).biHeight = height;
    (cast(BIH*)buffer[14..54]).biSizeImage = size_image;

    for (int y=0, i=0, add=0; y<height; ++y, i+=width_align, add+=pitch) {
        data[i..i+width_byte] = pixel[add..add+width_byte];
    }
    
    try std.file.write(nameFile, buffer);
    catch (Exception ex) throw new Exception("writeBitmap(); " ~ nameFile ~ " failed to create");
    
    return buffer;
}




// set bitmap file header
void setBitmapFile(inout ubyte[] header, uint width, uint height, uint bit)
{
    int width_bit  = width * bit;
    int width_align = ((width_bit>>3) + 3) & 0xfffffc;
    int size_image  = width_align * height;
    
    (cast(BFH*)header[0..14]).bfType[] = "BM";
    (cast(BFH*)header[0..14]).bfReserved1 = 0;
    (cast(BFH*)header[0..14]).bfReserved2 = 0;
    (cast(BFH*)header[0..14]).bfOffBits = 54;
    (cast(BFH*)header[0..14]).bfSize = size_image + 54;    
}




// set bitmap info header
void setBitmapInfo(inout ubyte[] header, uint width, uint height, uint bit)
{
    int width_bit  = width * bit;
    int width_align = ((width_bit>>3) + 3) & 0xfffffc;
    int size_image  = width_align * height;
    
    (cast(BIH*)header[0..40]).biSize = 40;
    (cast(BIH*)header[0..40]).biPlanes = 1;
    (cast(BIH*)header[0..40]).biCompression = 0;
    (cast(BIH*)header[0..40]).biXPelsPerMeter = 0;
    (cast(BIH*)header[0..40]).biYPelsPerMeter = 0;
    (cast(BIH*)header[0..40]).biClrUsed = 0;
    (cast(BIH*)header[0..40]).biClrImportant = 0;
    
    (cast(BIH*)header[0..40]).biWidth = width;
    (cast(BIH*)header[0..40]).biHeight = height;
    (cast(BIH*)header[0..40]).biBitCount = bit;
    (cast(BIH*)header[0..40]).biSizeImage = size_image;
}




// convert gray scale
ubyte[] convertBitmap8(out ubyte[] pix8, ubyte[] pixel, ubyte[] tableColor, int bit)
{
    uint  add, i;
    ubyte f;
    ubyte[256] gray;
    
    switch (bit) {
    case 24:
        pix8.length = pixel.length / 3;
        for (i=0, add=0; i<pix8.length; i++, add+=3) {
            pix8[i] = rgb2gray(pixel[add..add+3]);
        }
        return pix8;
    case 8:
        pix8.length = pixel.length;
        for (i=0; i<tableColor.length; i+=4) {
            gray[i>>2] = rgb2gray(tableColor[i..i+3]);
        }
        add=0;
        foreach(p; pixel) { 
            pix8[add] = gray[p];
            add++;
        }
        return pix8;
    case 1:
        pix8.length = pixel.length * 8;
        add = 0;
        gray[0] = rgb2gray(tableColor[0..3]);
        gray[1] = rgb2gray(tableColor[3..6]);
        foreach(p; pixel) {
            for (i=0; i<8; i++, p<<=1) {
                f=(p&128)>>7;
                pix8[add++] = gray[f];
            }
        }
        return pix8;
    default:
        throw new Exception("convertBitmap(); bit size not supported");
    }
}




// convert 24bit
ubyte[] convertBitmap24(out ubyte[] pix24, ubyte[] pixel, ubyte[] tableColor, int bit, bool isBGR)
{
    uint  add, i;
    ubyte f;
    
    ubyte[] function(ubyte[], ubyte[]) copybits = (isBGR) ? (&copyLE3) : (&copyBE3);
    
    switch (bit) {
    case 24:
        pix24.length = pixel.length;
        for (i=0; i<pixel.length; i+=3) {
            copybits(pix24[i..i+3], pixel[i..i+3]);
        }
        return pix24;
    case 8:
        pix24.length = pixel.length * 3;
        add = 0;
        foreach(p; pixel) {
            copybits(pix24[add..add+3], tableColor[p<<2..(p<<2)+3]);
            add+=3;
        }
        return pix24;
    case 1:
        pix24.length = pixel.length * 24;
        add = 0;
        foreach(p; pixel) {
            for (i=0; i<8; i++, add+=3, p<<=1) {
                f = (p&128)>>5;
                copybits(pix24[add..add+3], tableColor[f..f+3]);
            }
        }
        return pix24;
    default:
        throw new Exception("convertBitmap(); bit size not supported");
    }
}




// convert 24bit + alpha
ubyte[] convertBitmap32(out ubyte[] pix32, ubyte[] pixel, ubyte[] tableColor, int bit, bool isBGRA)
{
    uint  add, i;
    ubyte f;
    
    ubyte[] function(ubyte[], ubyte[]) copybits = (isBGRA) ? (&copyLE3) : (&copyBE3);

    switch (bit) {
    case 24:
        pix32.length = (pixel.length * 4) / 3;
        add = 0;
        for (i=0; i<pix32.length; i+=4, add+=3) {
            copybits(pix32[i..i+3], pixel[add..add+3]);
            pix32[i+3] = 255;
        }
        return pix32;
    case 8:
        pix32.length = pixel.length * 4;
        add = 0;
        foreach(p; pixel) {
            copybits(pix32[add..add+3], tableColor[p<<2..(p<<2)+3]);
            pix32[add+3] = 255; //tableColor[p+3];
            add+=4;
        }
        return pix32;
    case 1:
        pix32.length = pixel.length * 32;
        add = 0;
        foreach(p; pixel) {
            for (i=0; i<8; i++, add+=4, p<<=1) {
                f = (p&128)>>5;
                copybits(pix32[add..add+3], tableColor[f..f+3]);
                pix32[add+3] = 255; //tableColor[f+3];
            }
        }
        return pix32;
    default:
        throw new Exception("convertBitmap(); bit size not supported");
    }
}




// alpha color
ubyte[] transparentAlphaColor(inout ubyte[] pix32, uint colTrans)
{
    colTrans |= 0xff000000;
    for (int i=0; i<pix32.length; i+=4) {
        pix32[i+3] = (rgb2color(pix32[i..i+3]) == colTrans) ? 0 : 255;
    }
    return pix32;
}




// combine alpha map
ubyte[] setAlphaMap(inout ubyte[] pix32, ubyte[] pix8_alpha)
{
    for (int i=0; i<pix8_alpha.length; i++) { 
        pix32[(i<<2)+3] = pix8_alpha[i];
    }
    return pix32;
}




// combine alpha map with wrapping around
ubyte[] setAlphaMap(inout ubyte[] pix32, ubyte[] pix8_alpha, int width_pix32, int width_pix8)
{
    int x32, x8, y32, y8;
    width_pix32 <<= 2;
    for (y32=0, y8=0; y32<pix32.length; y32+=width_pix32) {
        for (x32=3, x8=0; x32<width_pix32; x32+=4) {
            pix32[y32+x32] = pix8_alpha[y8+x8];
            if (++x8 == width_pix8) { x8=0; } // wrap around
        }
        if ((y8+=width_pix8) >= pix8_alpha.length) { y8=0; } // wrap around
    }
    return pix32;
}

